@@ -2,7 +2,7 @@
 /********************************************/
 /* gd interface to freetype library         */
 /*                                          */
-/* John Ellson   ellson@lucent.com          */
+/* John Ellson   ellson@graphviz.org        */
 /********************************************/
 
 #ifdef HAVE_CONFIG_H
@@ -21,13 +21,13 @@
 #include <unistd.h>
 #else
 #include <io.h>
+#define R_OK 04			/* Needed in Windows */
 #endif
 
 /* number of antialised colors for indexed bitmaps */
 #define NUMCOLORS 8
 
-char *
-gdImageStringTTF (gdImage * im, int *brect, int fg, char *fontlist,
+BGD_DECLARE(char *) gdImageStringTTF (gdImage * im, int *brect, int fg, char *fontlist,
 		  double ptsize, double angle, int x, int y, char *string)
 {
   /* 2.0.6: valid return */
@@ -36,16 +36,14 @@ gdImageStringTTF (gdImage * im, int *brect, int fg, char *fontlist,
 }
 
 #ifndef HAVE_LIBFREETYPE
-char *
-gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
+BGD_DECLARE(char *) gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
 		   double ptsize, double angle, int x, int y, char *string,
 		   gdFTStringExtraPtr strex)
 {
   return "libgd was not built with FreeType font support\n";
 }
 
-char *
-gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
+BGD_DECLARE(char *) gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
 		 double ptsize, double angle, int x, int y, char *string)
 {
   return "libgd was not built with FreeType font support\n";
@@ -53,8 +51,11 @@ gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
 #else
 
 #include "gdcache.h"
-#include "freetype/freetype.h"
-#include "freetype/ftglyph.h"
+/* 2.0.16 Christophe Thomas: starting with FreeType 2.1.6, this is
+  mandatory, and it has been supported for a long while. */
+#include <ft2build.h>
+#include FT_FREETYPE_H
+#include FT_GLYPH_H
 
 /* number of fonts cached before least recently used is replaced */
 #define FONTCACHESIZE 6
@@ -71,19 +72,24 @@ gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
 
 /*
  * The character (space) used to separate alternate fonts in the
- * fontlist parameter to gdImageStringFT.
+ * fontlist parameter to gdImageStringFT. 2.0.18: space was a
+ * poor choice for this.
  */
-#define LISTSEPARATOR " "
+#define LISTSEPARATOR ";"
 
 /*
  * DEFAULT_FONTPATH and PATHSEPARATOR are host type dependent and
- * are normally set by configure in gvconfig.h.  These are just
+ * are normally set by configure in config.h.  These are just
  * some last resort values that might match some Un*x system
  * if building this version of gd separate from graphviz.
  */
 #ifndef DEFAULT_FONTPATH
+#if defined(__APPLE__) || (defined(__MWERKS__) && defined(macintosh))
+#define DEFAULT_FONTPATH "/usr/share/fonts/truetype:/System/Library/Fonts:/Library/Fonts"
+#else
 #define DEFAULT_FONTPATH "/usr/share/fonts/truetype"
 #endif
+#endif
 #ifndef PATHSEPARATOR
 #define PATHSEPARATOR ":"
 #endif
@@ -377,10 +383,11 @@ fontFetch (char **error, void *key)
        * Allocate an oversized buffer that is guaranteed to be
        * big enough for all paths to be tested.
        */
+      /* 2.0.22: Thorben Kundinger: +8 is needed, not +6. */
       fullname = gdRealloc (fullname,
-			    strlen (fontsearchpath) + strlen (name) + 6);
-      /* if name is an absolute filename then test directly */
-      if (*name == '/'
+			    strlen (fontsearchpath) + strlen (name) + 8);
+      /* if name is an absolute or relative pathname then test directly */
+      if (strchr (name, '/')
 	  || (name[0] != 0 && name[1] == ':'
 	      && (name[2] == '/' || name[2] == '\\')))
 	{
@@ -388,6 +395,8 @@ fontFetch (char **error, void *key)
 	  if (access (fullname, R_OK) == 0)
 	    {
 	      font_found++;
+	      /* 2.0.16: memory leak fixed, Gustavo Scotti */
+	      gdFree (path);
 	      break;
 	    }
 	}
@@ -412,6 +421,12 @@ fontFetch (char **error, void *key)
 	      font_found++;
 	      break;
 	    }
+	  sprintf (fullname, "%s/%s.dfont", dir, name);
+	  if (access (fullname, R_OK) == 0)
+	    {
+	      font_found++;
+	      break;
+	    }
 	}
       gdFree (path);
       if (font_found)
@@ -421,18 +436,36 @@ fontFetch (char **error, void *key)
   if (!font_found)
     {
       *error = "Could not find/open font";
+      /* 2.0.12: TBB: free these. Thanks to Frank Faubert. */
+      free (a->fontlist);
+      free (fullname);
+      gdFree (a);
+
       return NULL;
     }
 
   err = FT_New_Face (*b->library, fullname, 0, &a->face);
   if (err)
     {
+      /* 2.0.12: TBB: free these. Thanks to Frank Faubert. */
+      free (a->fontlist);
+      free (fullname);
+      gdFree (a);
       *error = "Could not read font";
       return NULL;
     }
   gdFree (fullname);
 
 /* FIXME - This mapping stuff is imcomplete - where is the spec? */
+/* EAM   - It's worse than that. It's pointless to match character encodings here. */
+/*         As currently written, the stored a->face->charmap only matches one of   */
+/*         the actual charmaps and we cannot know at this stage if it is the right */
+/*         one. We should just skip all this stuff, and check in gdImageStringFTEx */
+/*         if some particular charmap is preferred and if so whether it is held in */
+/*         one of the a->face->charmaps[0..num_charmaps].                          */
+/*         And why is it so bad not to find any recognized charmap?  The user may  */
+/*         still know what mapping to use, even if we do not.  In that case we can */
+/*         just use the map in a->face->charmaps[num_charmaps] and be done with it.*/
 
   a->have_char_map_unicode = 0;
   a->have_char_map_big5 = 0;
@@ -443,6 +476,24 @@ fontFetch (char **error, void *key)
       charmap = a->face->charmaps[n];
       platform = charmap->platform_id;
       encoding = charmap->encoding_id;
+
+/* EAM DEBUG - Newer versions of libfree2 make it easier 
+	by defining encodings */
+/* TBB: get this exactly right: 2.1.3 *or better*, all possible cases. */
+
+#if ((defined(FREETYPE_MAJOR)) && (((FREETYPE_MAJOR == 2) && (((FREETYPE_MINOR == 1) && (FREETYPE_PATCH >= 3)) || (FREETYPE_MINOR > 1))) || (FREETYPE_MAJOR > 2)))
+      if (charmap->encoding == FT_ENCODING_MS_SYMBOL
+	  || charmap->encoding == FT_ENCODING_ADOBE_CUSTOM
+	  || charmap->encoding == FT_ENCODING_ADOBE_STANDARD)
+	{
+	  a->have_char_map_unicode = 1;
+	  found = charmap;
+	  a->face->charmap = charmap;
+	  return (void *) a;
+	}
+#endif /* Freetype 2.1 or better */
+/* EAM DEBUG */
+
       if ((platform == 3 && encoding == 1)	/* Windows Unicode */
 	  || (platform == 3 && encoding == 0)	/* Windows Symbol */
 	  || (platform == 2 && encoding == 1)	/* ISO Unicode */
@@ -471,6 +522,9 @@ fontFetch (char **error, void *key)
   if (!found)
     {
       *error = "Unable to find a CharMap that I can handle";
+      /* 2.0.12: TBB: free these. Thanks to Frank Faubert. */
+      free (a->fontlist);
+      gdFree (a);
       return NULL;
     }
   /* 2.0.5: we should actually return this */
@@ -597,7 +651,8 @@ gdft_draw_bitmap (gdCache_head_t * tc_cache, gdImage * im, int fg,
 	  pcr = pc;
 	  y = pen_y + row;
 	  /* clip if out of bounds */
-	  if (y >= im->sy || y < 0)
+	  /* 2.0.16: clipping rectangle, not image bounds */
+	  if ((y > im->cy2) || (y < im->cy1))
 	    continue;
 	  for (col = 0; col < bitmap.width; col++, pc++)
 	    {
@@ -637,7 +692,8 @@ gdft_draw_bitmap (gdCache_head_t * tc_cache, gdImage * im, int fg,
 	      level = gdAlphaMax - level;
 	      x = pen_x + col;
 	      /* clip if out of bounds */
-	      if (x >= im->sx || x < 0)
+	      /* 2.0.16: clip to clipping rectangle, Matt McNabb */
+	      if ((x > im->cx2) || (x < im->cx1))
 		continue;
 	      /* get pixel location in gd buffer */
 	      tpixel = &im->tpixels[y][x];
@@ -747,33 +803,59 @@ gdroundupdown (FT_F26Dot6 v1, int updown)
 extern int any2eucjp (char *, char *, unsigned int);
 
 /* Persistent font cache until explicitly cleared */
-/*     Fonts can be used across multiple images */
+/* Fonts can be used across multiple images */
+
+/* 2.0.16: thread safety (the font cache is shared) */
+gdMutexDeclare (gdFontCacheMutex);
 static gdCache_head_t *fontCache;
 static FT_Library library;
 
-void
-gdFreeFontCache ()
+BGD_DECLARE(void) gdFreeFontCache ()
+{
+  gdFontCacheShutdown ();
+}
+
+BGD_DECLARE(void) gdFontCacheShutdown ()
 {
   if (fontCache)
     {
+      gdMutexShutdown (gdFontCacheMutex);
       gdCacheDelete (fontCache);
       FT_Done_FreeType (library);
+      /* 2.0.16: Gustavo Scotti: make sure we don't free this twice */
+      fontCache = 0;
     }
 }
 
 /********************************************************************/
 /* gdImageStringFT -  render a utf8 string onto a gd image          */
 
-char *
-gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
+BGD_DECLARE(char *) gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
 		 double ptsize, double angle, int x, int y, char *string)
 {
   return gdImageStringFTEx (im, brect, fg, fontlist,
 			    ptsize, angle, x, y, string, 0);
 }
 
-char *
-gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
+BGD_DECLARE(int) gdFontCacheSetup (void)
+{
+  if (fontCache)
+    {
+      /* Already set up */
+      return 0;
+    }
+  gdMutexSetup (gdFontCacheMutex);
+  if (FT_Init_FreeType (&library))
+    {
+      gdMutexShutdown (gdFontCacheMutex);
+      return -1;
+    }
+  fontCache = gdCacheCreate (FONTCACHESIZE, fontTest, fontFetch, fontRelease);
+  return 0;
+}
+
+
+BGD_DECLARE(char *) gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
 		   double ptsize, double angle, int x, int y, char *string,
 		   gdFTStringExtraPtr strex)
 {
@@ -796,7 +878,10 @@ gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
   char *tmpstr = 0;
   int render = (im && (im->trueColor || (fg <= 255 && fg >= -255)));
   FT_BitmapGlyph bm;
-  int render_mode = FT_LOAD_RENDER | FT_LOAD_FORCE_AUTOHINT;
+  /* 2.0.13: Bob Ostermann: don't force autohint, that's just for testing 
+     freetype and doesn't look as good */
+  int render_mode = FT_LOAD_DEFAULT;
+  int m, mfound;
   /* Now tuneable thanks to Wez Furlong */
   double linespace = LINESPACE;
   /* 2.0.6: put this declaration with the other declarations! */
@@ -808,7 +893,8 @@ gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
    *   colorindexes.          -- 27.06.2001 <krisku@arrak.fi>
    */
   gdCache_head_t *tc_cache;
-
+  /* Tuneable horizontal and vertical resolution in dots per inch */
+  int hdpi, vdpi;
   if (strex)
     {
       if ((strex->flags & gdFTEX_LINESPACE) == gdFTEX_LINESPACE)
@@ -821,19 +907,16 @@ gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
 			    tweenColorRelease);
 
 /***** initialize font library and font cache on first call ******/
-
   if (!fontCache)
     {
-      if (FT_Init_FreeType (&library))
+      if (gdFontCacheSetup () != 0)
 	{
 	  gdCacheDelete (tc_cache);
 	  return "Failure to initialize font library";
 	}
-      fontCache = gdCacheCreate (FONTCACHESIZE,
-				 fontTest, fontFetch, fontRelease);
     }
 /*****/
-
+  gdMutexLock (gdFontCacheMutex);
   /* get the font (via font cache) */
   fontkey.fontlist = fontlist;
   fontkey.library = &library;
@@ -841,15 +924,29 @@ gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
   if (!font)
     {
       gdCacheDelete (tc_cache);
+      gdMutexUnlock (gdFontCacheMutex);
       return fontCache->error;
     }
   face = font->face;		/* shortcut */
   slot = face->glyph;		/* shortcut */
 
-  if (FT_Set_Char_Size (face, 0, (FT_F26Dot6) (ptsize * 64),
-			GD_RESOLUTION, GD_RESOLUTION))
+  /*
+   * Added hdpi and vdpi to support images at non-screen resolutions, i.e. 300 dpi TIFF,
+   *    or 100h x 50v dpi FAX format. 2.0.23.
+   * 2004/02/27 Mark Shackelford, mark.shackelford@acs-inc.com
+   */
+  hdpi = GD_RESOLUTION;
+  vdpi = GD_RESOLUTION;
+  if (strex && (strex->flags & gdFTEX_RESOLUTION))
+    {
+      hdpi = strex->hdpi;
+      vdpi = strex->vdpi;
+    }
+
+  if (FT_Set_Char_Size (face, 0, (FT_F26Dot6) (ptsize * 64), hdpi, vdpi))
     {
       gdCacheDelete (tc_cache);
+      gdMutexUnlock (gdFontCacheMutex);
       return "Could not set character size";
     }
 
@@ -868,6 +965,51 @@ gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
     {
       render_mode |= FT_LOAD_MONOCHROME;
     }
+  /* 2.0.12: allow explicit specification of the preferred map;
+     but we still fall back if it is not available. */
+  m = gdFTEX_Unicode;
+  if (strex && (strex->flags & gdFTEX_CHARMAP))
+    {
+      m = strex->charmap;
+    }
+  /* Try all three types of maps, but start with the specified one */
+  mfound = 0;
+  for (i = 0; (i < 3); i++)
+    {
+      switch (m)
+	{
+	case gdFTEX_Unicode:
+	  if (font->have_char_map_unicode)
+	    {
+	      mfound = 1;
+	    }
+	  break;
+	case gdFTEX_Shift_JIS:
+	  if (font->have_char_map_sjis)
+	    {
+	      mfound = 1;
+	    }
+	  break;
+	case gdFTEX_Big5:
+	  {
+	    /* This was the 'else' case, we can't really 'detect' it */
+	    mfound = 1;
+	  }
+	  break;
+	}
+      if (mfound)
+	{
+	  break;
+	}
+      m++;
+      m %= 3;
+    }
+  if (!mfound)
+    {
+      /* No character set found! */
+      gdMutexUnlock (gdFontCacheMutex);
+      return "No character set found";
+    }
 
 #ifndef JISX0208
   if (font->have_char_map_sjis)
@@ -907,6 +1049,8 @@ gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
       /* newlines */
       if (ch == '\n')
 	{
+	  /* 2.0.13: reset penf.x. Christopher J. Grayce */
+	  penf.x = 0;
 	  penf.y -= face->size->metrics.height * linespace;
 	  penf.y = (penf.y - 32) & -64;	/* round to next pixel row */
 	  x1 = (penf.x * cos_a - penf.y * sin_a + 32) / 64;
@@ -917,62 +1061,87 @@ gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
 	  continue;
 	}
 
-      if (font->have_char_map_unicode)
+/* EAM DEBUG */
+/* TBB: get this exactly right: 2.1.3 *or better*, all possible cases. */
+/* 2.0.24: David R. Morrison: use the more complete ifdef here. */
+
+#if ((defined(FREETYPE_MAJOR)) && (((FREETYPE_MAJOR == 2) && (((FREETYPE_MINOR == 1) && (FREETYPE_PATCH >= 3)) || (FREETYPE_MINOR > 1))) || (FREETYPE_MAJOR > 2)))
+      if (font->face->charmap->encoding == FT_ENCODING_MS_SYMBOL)
 	{
-	  /* use UTF-8 mapping from ASCII */
+	  /* I do not know the significance of the constant 0xf000. */
+	  /* It was determined by inspection of the character codes */
+	  /* stored in Microsoft font symbol.ttf                    */
 	  len = gdTcl_UtfToUniChar (next, &ch);
+	  ch |= 0xf000;
 	  next += len;
 	}
-      else if (font->have_char_map_sjis)
-	{
-	  unsigned char c;
-	  int jiscode;
-
-	  c = *next;
-	  if (0xA1 <= c && c <= 0xFE)
-	    {
-	      next++;
-	      jiscode = 0x100 * (c & 0x7F) + ((*next) & 0x7F);
-
-	      ch = (jiscode >> 8) & 0xFF;
-	      jiscode &= 0xFF;
-
-	      if (ch & 1)
-		jiscode += 0x40 - 0x21;
-	      else
-		jiscode += 0x9E - 0x21;
-
-	      if (jiscode >= 0x7F)
-		jiscode++;
-	      ch = (ch - 0x21) / 2 + 0x81;
-	      if (ch >= 0xA0)
-		ch += 0x40;
-
-	      ch = (ch << 8) + jiscode;
-	    }
-	  else
-	    {
-	      ch = c & 0xFF;	/* don't extend sign */
-	    }
-	  next++;
-	}
       else
-	{
-	  /*
-	   * Big 5 mapping:
-	   * use "JIS-8 half-width katakana" coding from 8-bit characters. Ref:
-	   * ftp://ftp.ora.com/pub/examples/nutshell/ujip/doc/japan.inf-032092.sjs
-	   */
-	  ch = (*next) & 0xFF;	/* don't extend sign */
-	  next++;
-	  if (ch >= 161		/* first code of JIS-8 pair */
-	      && *next)
-	    {			/* don't advance past '\0' */
-	      /* TBB: Fix from Kwok Wah On: & 255 needed */
-	      ch = (ch * 256) + ((*next) & 255);
+#endif /* Freetype 2.1 or better */
+/* EAM DEBUG */
+
+	switch (m)
+	  {
+	  case gdFTEX_Unicode:
+	    if (font->have_char_map_unicode)
+	      {
+		/* use UTF-8 mapping from ASCII */
+		len = gdTcl_UtfToUniChar (next, &ch);
+		next += len;
+	      }
+	    break;
+	  case gdFTEX_Shift_JIS:
+	    if (font->have_char_map_sjis)
+	      {
+		unsigned char c;
+		int jiscode;
+		c = *next;
+		if (0xA1 <= c && c <= 0xFE)
+		  {
+		    next++;
+		    jiscode = 0x100 * (c & 0x7F) + ((*next) & 0x7F);
+
+		    ch = (jiscode >> 8) & 0xFF;
+		    jiscode &= 0xFF;
+
+		    if (ch & 1)
+		      jiscode += 0x40 - 0x21;
+		    else
+		      jiscode += 0x9E - 0x21;
+
+		    if (jiscode >= 0x7F)
+		      jiscode++;
+		    ch = (ch - 0x21) / 2 + 0x81;
+		    if (ch >= 0xA0)
+		      ch += 0x40;
+
+		    ch = (ch << 8) + jiscode;
+		  }
+		else
+		  {
+		    ch = c & 0xFF;	/* don't extend sign */
+		  }
+		next++;
+	      }
+	    break;
+	  case gdFTEX_Big5:
+	    {
+	      /*
+	       * Big 5 mapping:
+	       * use "JIS-8 half-width katakana" coding from 8-bit characters. Ref:
+	       * ftp://ftp.ora.com/pub/examples/nutshell/ujip/doc/japan.inf-032092.sjs
+	       */
+	      ch = (*next) & 0xFF;	/* don't extend sign */
 	      next++;
+	      if (ch >= 161	/* first code of JIS-8 pair */
+		  && *next)
+		{		/* don't advance past '\0' */
+		  /* TBB: Fix from Kwok Wah On: & 255 needed */
+		  ch = (ch * 256) + ((*next) & 255);
+		  next++;
+		}
 	    }
-	}
+	    break;
+	  }
       /* set rotation transform */
       FT_Set_Transform (face, &matrix, NULL);
       /* Convert character code to glyph index */
@@ -984,6 +1153,8 @@ gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
 	  FT_Get_Kerning (face, previous, glyph_index,
 			  ft_kerning_default, &delta);
 	  pen.x += delta.x;
+	  /* 2.0.12: indispensable for an accurate bounding box. */
+	  penf.x += delta.x;
 	}
 
       /* load glyph image into the slot (erase previous one) */
@@ -991,6 +1162,7 @@ gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
       if (err)
 	{
 	  gdCacheDelete (tc_cache);
+	  gdMutexUnlock (gdFontCacheMutex);
 	  return "Problem loading glyph";
 	}
 
@@ -1034,6 +1206,7 @@ gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
 	      if (err)
 		{
 		  gdCacheDelete (tc_cache);
+		  gdMutexUnlock (gdFontCacheMutex);
 		  return "Problem rendering glyph";
 		}
 	    }
@@ -1087,6 +1260,7 @@ gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
   if (tmpstr)
     gdFree (tmpstr);
   gdCacheDelete (tc_cache);
+  gdMutexUnlock (gdFontCacheMutex);
   return (char *) NULL;
 }
 
